package chocopy.pa3;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import chocopy.common.analysis.AbstractNodeAnalyzer;
import chocopy.common.analysis.SymbolTable;
import chocopy.common.astnodes.*;
import chocopy.common.analysis.types.*;
import chocopy.common.codegen.*;

import static chocopy.common.codegen.RiscVBackend.Register.*;
import chocopy.common.codegen.RiscVBackend.Register;


/**
 * This is where the main implementation of PA3 will live.
 *
 * <p>
 * A large part of the functionality has already been implemented in the base
 * class, CodeGenBase. Make sure to read through that class, since you will want
 * to use many of its fields and utility methods in this class when emitting
 * code.
 *
 * <p>
 * Also read the PDF spec for details on what the base class does and what APIs
 * it exposes for its sub-class (this one). Of particular importance is knowing
 * what all the SymbolInfo classes contain.
 */
public class CodeGenImpl extends CodeGenBase 
{

    /** A code generator emitting instructions to BACKEND. */
    public CodeGenImpl(RiscVBackend backend) 
    {
        super(backend);
        
    }
    
    /**
     * Generates assembly code for PROGRAM.
     *
     * <p>This is the main driver that calls internal methods for emitting DATA section (globals,
     * constants, prototypes, etc) as well as the the CODE section (predefined functions, built-in
     * routines, and user-defined functions).
     */
    @Override
    public void generate(Program program) {
    		System.out.println("Inside generate");   
        analyzeProgram(program);

        backend.startData();

        for (ClassInfo classInfo : this.classes) {
            emitPrototype(classInfo);
        }

        for (ClassInfo classInfo : this.classes) {
            emitDispatchTable(classInfo);
        }

        for (GlobalVarInfo global : this.globalVars) {
            backend.emitGlobalLabel(global.getLabel());
            emitConstant(
                    global.getInitialValue(),
                    global.getVarType(),
                    String.format("Initial value of global var: %s", global.getVarName()));
        }

        backend.startCode();

        Label mainLabel = new Label("main");
        backend.emitGlobalLabel(mainLabel);
        backend.emitLUI(A0, HEAP_SIZE_BYTES >> 12, "Initialize heap size (in multiples of 4KB)");
        backend.emitADD(S11, S11, A0, "Save heap size");
        backend.emitJAL(heapInitLabel, "Call heap.init routine");
        backend.emitMV(GP, A0, "Initialize heap pointer");
        backend.emitMV(S10, GP, "Set beginning of heap");
        backend.emitADD(S11, S10, S11, "Set end of heap (= start of heap + heap size)");
        backend.emitMV(RA, ZERO, "No normal return from main program.");
        backend.emitMV(FP, ZERO, "No preceding frame.");

        emitTopLevel(program.statements);

        for (FuncInfo funcInfo : this.functions) {
            funcInfo.emitBody();
        }

        emitStdFunc("alloc");
        emitStdFunc("alloc2");
        emitStdFunc("abort");
        emitStdFunc("heap.init");
        emitStdFunc("makeint");

        emitCustomCode();

        backend.startData();
        emitConstants();
    }

    /** Operation on None. */
    private final Label errorNone = new Label("error.None");
    /** Division by zero. */
    private final Label errorDiv = new Label("error.Div");
    /** Index out of bounds. */
    private final Label errorOob = new Label("error.OOB");
    
    /** Not implemented. */
    private final Label errorNI = new Label("error.NI");
    
    
    /** Label for built-in routine: makeint. */
    protected final Label makeIntLabel = new Label("makeint");
    
    
    /**
     * Emits the top level of the program.
     *
     * <p>
     * This method is invoked exactly once, and is surrounded by some boilerplate
     * code that: (1) initializes the heap before the top-level begins and (2) exits
     * after the top-level ends.
     *
     * <p>
     * You only need to generate code for statements.
     *
     * @param statements top level statements
     */
    protected void emitTopLevel(List<Stmt> statements) 
    {
        StmtAnalyzer stmtAnalyzer = new StmtAnalyzer(null);
        backend.emitADDI(SP, SP, -2 * backend.getWordSize(), "Saved FP and saved RA (unused at top level).");
        backend.emitSW(ZERO, SP, 0, "Top saved FP is 0.");
        backend.emitSW(ZERO, SP, 4, "Top saved RA is 0.");
        backend.emitADDI(FP, SP, 2 * backend.getWordSize(), "Set FP to previous SP.");

        for (Stmt stmt : statements) 
        {
            stmt.dispatch(stmtAnalyzer);
        }
        backend.emitLI(A0, EXIT_ECALL, "Code for ecall: exit");
        backend.emitEcall(null);
    }

    /**
     * Emits the code for a function described by FUNCINFO.
     *
     * <p>
     * This method is invoked once per function and method definition. At the code
     * generation stage, nested functions are emitted as separate functions of their
     * own. So if function `bar` is nested within function `foo`, you only emit
     * `foo`'s code for `foo` and only emit `bar`'s code for `bar`.
     */
    protected void emitUserDefinedFunction(FuncInfo funcInfo) 
    {
        backend.emitGlobalLabel(funcInfo.getCodeLabel());
        
        
        
        // --- Prologue ---
        // space for return address = 1
        // space for control link = 1
        // space for static link =  1
        // space for params = num of params 
        // space for locals = num of locals
        int requiredStackSpace = (1 + 1 + 1 + funcInfo.getLocals().size() + funcInfo.getParams().size())*wordSize;
        
        backend.emitADDI(SP, SP, -requiredStackSpace, "Reserve space for stack frame.");
        backend.emitSW(RA, SP, requiredStackSpace-4, "return address");
        backend.emitSW(FP, SP, requiredStackSpace-8, "control link");
        // if we want to add static link
        //backend.emitSW(FP, SP, requiredStackSpace-12, "static link");
        backend.emitADDI(FP, SP, requiredStackSpace, "New fp is at old SP.");
        
        
        StmtAnalyzer stmtAnalyzer = new StmtAnalyzer(funcInfo);
        int emptySlotNum = 4;
        for (StackVarInfo var : funcInfo.getLocals()) {
    			// store at requiredStackSpace-emptySlotNum*wordSize(SP), then emptySlotNum++
        		Literal varLitral = var.getInitialValue();
        		varLitral.dispatch(stmtAnalyzer);
        		// All Literals should save locations for the values in A0
    			backend.emitSW(A0, SP, requiredStackSpace-emptySlotNum*wordSize, "Push local variable " + var.getVarName() + " onto stack");
    			emptySlotNum++;
        }
        
        // --- Function Body ---
        // statements load all the variables that caller put on stack
        // statements use fp to load the variables
        // example: 0(fp) is the last variable (z) while 8(fp) is the first variable (x) 
        // for function with 3 params f(x, y, z)
        for (Stmt stmt : funcInfo.getStatements()) {
        	    //System.out.println(stmt.toString());
            stmt.dispatch(stmtAnalyzer);
        }
        backend.emitJ(stmtAnalyzer.epilogue, "Jump to function epilogue");

        
        
        // --- Epilogue ---
        backend.emitLocalLabel(stmtAnalyzer.epilogue, "Epilogue");
        backend.emitLW(RA, FP, -4, "Get return address");
        backend.emitLW(FP, FP, -8, "Use control link to restore caller's fp");
        backend.emitADDI(SP, SP, requiredStackSpace, "Restore stack pointer");
        backend.emitJR(RA, "Return to caller");
    }

    /** An analyzer that encapsulates code generation for statements. */
    private class StmtAnalyzer extends AbstractNodeAnalyzer<Register> {
        /*
         * The symbol table has all the info you need to determine what a given
         * identifier 'x' in the current scope is. You can use it as follows: SymbolInfo
         * x = sym.get("x");
         *
         * A SymbolInfo can be one the following: - ClassInfo: a descriptor for classes
         * - FuncInfo: a descriptor for functions/methods - AttrInfo: a descriptor for
         * attributes - GlobalVarInfo: a descriptor for global variables - StackVarInfo:
         * a descriptor for variables allocated on the stack, such as locals and
         * parameters
         *
         * Since the input program is assumed to be semantically valid and well-typed at
         * this stage, you can always assume that the symbol table contains valid
         * information. For example, in an expression `foo()` you KNOW that
         * sym.get("foo") will either be a FuncInfo or ClassInfo, but not any of the
         * other infos and never null.
         *
         * The symbol table in funcInfo has already been populated in the base class:
         * CodeGenBase. You do not need to add anything to the symbol table. Simply
         * query it with an identifier name to get a descriptor for a function, class,
         * variable, etc.
         *
         * The symbol table also maps nonlocal and global vars, so you only need to
         * lookup one symbol table and it will fetch the appropriate info for the var
         * that is currently in scope.
         */

        /** Symbol table for my statements. */
        private final SymbolTable<SymbolInfo> sym;

        /** Label of code that exits from procedure. */
        protected final Label epilogue;

        /** The descriptor for the current function, or null at the top level. */
        private final FuncInfo funcInfo;

        /** Label of code that exits from block. */
        protected Label elseBlock;

        /** Variable to keep track of offsets of stored variables */
        private Map<SymbolInfo, Integer> offsetMap = new HashMap<>();
        private final String size_label;
        /** Variable to store offset from frame pointer to identify next 
          * empty space on stack frame to store variable*/
        private int sp_off;
        /** Variable to store maximum possible offset depending on stack size.*/
        private int max_sp;
        
        /**
         * An analyzer for the function described by FUNCINFO0, which is null for the
         * top level.
         */
        StmtAnalyzer(FuncInfo funcInfo0) {
            funcInfo = funcInfo0;
            if (funcInfo == null) {
                sym = globalSymbols;
                sp_off = max_sp = 2;
                size_label = "@..main.size";
            } else {
                sym = funcInfo.getSymbolTable();
                sp_off = max_sp = funcInfo0.getLocals().size() + 2;
                size_label = "@"+funcInfo0.getFuncName()+".size";
            }
            epilogue = generateLocalLabel();
        }
        private void incSp(int i){
            sp_off+=i+1;
            max_sp = max_sp >= sp_off?max_sp:sp_off;
        }
        
        
        
        
        // *********** functions start ***********
        
        public Register analyze(CallExpr node) {
        		backend.emitLW(T6, FP, 0, "Inside CallExpr: " + node.function.name);
            SymbolInfo Ty = globalSymbols.get(node.function.name);
            if(Ty instanceof ClassInfo){
                //object create
                ClassInfo cls = (ClassInfo) Ty;
                /**
                    la a0, $DoublingVector$prototype         # Load pointer to prototype of: DoublingVector
                    jal alloc                                # Allocate new object in A0
                    sw a0, -12(fp)                           # Push on stack slot 3
                    sw a0, -16(fp)                           # Push argument 0 from last.
                    addi sp, fp, -16                         # Set SP to last argument.
                    lw a1, 8(a0)                             # Load address of object's dispatch table
                    lw a1, 0(a1)                             # Load address of method: DoublingVector.__init__
                    jalr a1                                  # Invoke method: DoublingVector.__init__
                    addi sp, fp, -@..main.size               # Set SP to stack frame top.
                    lw a0, -12(fp)                           # Pop stack slot 3
                 */
                backend.emitLA(A0, cls.getPrototypeLabel(), 
                    String.format("Load pointer to prototype of: %s", cls.getClassName()));
                backend.emitJAL(objectAllocLabel, "Allocate new object in A0");
                backend.emitSW(A0, FP, -sp_off*wordSize, String.format("Push on stack slot %d", sp_off));
                incSp(0);

                backend.emitSW(A0, FP, -sp_off*wordSize, "Push argument 0 from last.");
                backend.emitADDI(SP, FP, sp_off, "Set SP to last argument.");
                backend.emitLW(A1, A0, getDispatchTableOffset(), "Load address of object's dispatch table");
                backend.emitLW(A1, A1, getMethodOffset(cls, "__init__"), String.format("Load address of method: %s.__init__", cls.getClassName()));
                backend.emitJALR(A1, String.format("Invoke method: %s.__init", cls.getClassName()));
                backend.emitADDI(SP, FP, "-"+size_label, "Set SP to stack frame top.");
                -- sp_off;
                backend.emitLW(A0, FP, -sp_off*wordSize, String.format("Pop stack slot %d", sp_off));
            } else {	    			
	        		// function
	        		Identifier functionId = node.function;
	        		
	        		List<Expr> args = node.args;
	        		int spaceRequiredForArgs = (args.size() + 1)*4;
	        		//backend.emitSW(A0, SP, -4, "Put static link");
	        		for (int i = 0; i < args.size(); i++) {
	        			int argNum = i + 1;
	        			int slotNum = argNum + 1; // We have extra slot for static link
	        			Expr expr = args.get(i);
	        			expr.dispatch(this);
	        			// All expressions should save their end result in A0
	        			// So, once expr is evaluated add value inside A0 onto stack as an argument
	        			backend.emitSW(A0, SP, -wordSize*slotNum, "Push argument " + argNum + " from left");
	        		}
	        		
	        		backend.emitADDI(SP, SP, -spaceRequiredForArgs, "Set SP to last argument.");
	        		backend.emitJAL(new Label("$"+functionId.name), "Invoke function: " + functionId.name);
	        		backend.emitADDI(SP, SP, spaceRequiredForArgs, "Set SP to stack frame top.");
            }
            return A0;
        }
        
        
      
      
      @Override
      public Register analyze(FuncDef node) {
    	  		backend.emitLW(T6, FP, 0, "Inside FuncDef: " + node.name);
      		// function
    	  		return null;
      }
      
      
      
      @Override
      public Register analyze(GlobalDecl node) {
    	  		backend.emitLW(T6, FP, 0, "Inside GlobalDecl: ");
      		// function
    	  		return null;
      }
      
      
      
      @Override
      public Register analyze(NonLocalDecl node) {
    	  		backend.emitLW(T6, FP, 0, "Inside NonLocalDecl: ");
      		// function
    	  		return null;
      }
      
      
        
        public Register analyze(MethodCallExpr node) {
        		backend.emitLW(T6, FP, 0, "Inside MethodCallExpr: " + node.method.member.name);
            Register obj = node.method.object.dispatch(this);
            int n_args = node.args.size();
            
            Label label = generateLocalLabel();
            backend.emitBNEZ(obj, label, "Ensure not None");
            backend.emitJ(errorNone, "Go to error handler");
            backend.emitLocalLabel(label, "Not None");
            incSp(n_args+1);
            backend.emitSW(obj, FP, (n_args - sp_off) *wordSize, String.format("Push argument %d from last.", n_args));

            for (int i = 0; i < n_args; ++i) 
                backend.emitSW(node.args.get(i).dispatch(this), FP, (n_args - i - 1 - sp_off) * wordSize, 
                    String.format("Push argument %d from last.", n_args - i - 1));
            
            backend.emitLW(A0, FP, (n_args- sp_off) * wordSize, String.format("Peek stack slot %d", sp_off - (n_args + 1)));
            ClassInfo objectClass = (ClassInfo)sym.get(((Identifier)node.method.object).name);

            backend.emitLW(A1, A0, getDispatchTableOffset(), "Load address of object's dispatch table");
            backend.emitLW(A1, A1, getMethodOffset(objectClass, node.method.member.name), 
                String.format("Load address of method: %s.%s", objectClass.getClassName(), node.method.member.name));
            backend.emitADDI(SP, FP, -sp_off * wordSize, "Set SP to last argument.");
            backend.emitJALR(A1, String.format("Invoke method: %s.%s", objectClass.getClassName(), node.method.member.name));
            backend.emitInsn(String.format("addi sp, fp, -%s", size_label), "Set SP to stack frame top.");
            sp_off -= n_args+1;
    
            return A0;
        }
        
        
        @Override
        public Register analyze(ReturnStmt stmt) {
        		backend.emitLW(T6, FP, 0, "Inside ReturnStmt: ");
        	    Expr expr = stmt.value;
        	    if (expr == null) {
        	    		backend.emitMV(A0, ZERO, "Return None");
        	    		return A0;
        	    }
        	    
        	    // All expressions should save their end result in A0
        	    expr.dispatch(this);
            return A0;
        }
        
        
        // *********** functions end ***********


        @Override
        public Register analyze(NoneLiteral node) 
        {
        		backend.emitLW(T6, FP, 0, "Inside NoneLiteral: ");
            backend.emitMV(Register.A0, Register.ZERO, "Load none");
            return Register.A0;
        }
        
        @Override
        public Register analyze(StringLiteral node) 
        {
        		backend.emitLW(T6, FP, 0, "Inside StringLiteral: ");
            Label l = constants.getStrConstant(node.value);
            backend.emitLA(Register.A0, l, "Load string literal");
            return Register.A0;
        }
        
        
// FIXME: This is not correct. We need to use $int$prototype to create an integer literal, so commenting it
//        @Override
//        public Register analyze(IntegerLiteral node) 
//        {
//            backend.emitLI(Register.A0, node.value, "Load integer literal "+node.value);
//            return Register.A0;
//        }
        
        @Override
        public Register analyze(BooleanLiteral node) 
        {
        		backend.emitLW(T6, FP, 0, "Inside BooleanLiteral: ");
            if(node.value==true)
                backend.emitLI(Register.A0, 1, "Load boolean literal: true ");
            else
                backend.emitLI(Register.A0, 0, "Load boolean literal: false ");
            return Register.A0;
        }
        
        @Override
        public Register analyze(AssignStmt node) 
        {
        		backend.emitLW(T6, FP, 0, "Inside AssignStmt: ");
            Type t = node.value.getInferredType();
            if(t.isSpecialType() || t.isListType())
            {
                node.value.dispatch(this);
                if (sym.getParent() == null) 
                {
                    for(Expr target: node.targets)
                    {
                        GlobalVarInfo gvi=(GlobalVarInfo)sym.get(((Identifier)target).name);
                        backend.emitSW(Register.A0, gvi.getLabel(), Register.T0, "Assign global: "+gvi.getVarName()+"(using tmp register)");
                    }
                }
                else
                {
                    for(Expr target: node.targets)
                    {
                        StackVarInfo svi = (StackVarInfo) sym.get(((Identifier)target).name);
                        int loc = offsetMap.get(svi);
                        backend.emitSW(Register.A0, Register.FP, -loc*4, "Load local variable: "+svi.getVarName());
                    }
                }
            }
            else
            {//TODO: Object Assignment

            }
            return Register.A0;
        }
        
        @Override
        public Register analyze(ExprStmt node) 
        {
        		backend.emitLW(T6, FP, 0, "Inside ExprStmt: ");
            node.expr.dispatch(this);
            return null;
        }
        
        @Override
        public Register analyze(IfExpr node) 
        {
            node.condition.dispatch(this);
            Label ln = generateLocalLabel();
            node.thenExpr.dispatch(this);
            backend.emitJ(ln, "Jump to end of if expression");
            backend.emitLocalLabel(elseBlock, "Else part of if expression");
            node.elseExpr.dispatch(this);
            backend.emitLocalLabel(ln, "End of if expression");
            return null;
        }
        
        @Override
        public Register analyze(IfStmt node) 
        {
            node.condition.dispatch(this);
            Label ln = generateLocalLabel();
            for(Stmt s:node.thenBody)
                s.dispatch(this);
            backend.emitJ(ln, "Jump to end of if statement");
            backend.emitLocalLabel(elseBlock, "Else part of if statement");
            for(Stmt s:node.elseBody)
                s.dispatch(this);
            backend.emitLocalLabel(ln, "End of if statement");
            return null;
        }
        
        @Override
        public Register analyze(BinaryExpr node) 
        {
            String operator = node.operator;
            backend.emitLW(T6, FP, 0, "Inside BinaryExpr: ");

            if(node.left.getInferredType().equals(Type.INT_TYPE) && node.right.getInferredType().equals(Type.INT_TYPE))
            {
                node.left.dispatch(this);
                backend.emitSW(Register.A0, Register.FP, -sp_off*wordSize, "Push on stack slot "+sp_off);
                sp_off++;
                node.right.dispatch(this);
                sp_off--;
                backend.emitLW(Register.T0, Register.FP, -sp_off*wordSize, "Pop stack slot "+sp_off);
            
                // Arithmetic Operators
                if(operator.equals("+"))
                    backend.emitADD(Register.A0, Register.A0, Register.T0, "Add operation");
                else if(operator.equals("-"))
                    backend.emitSUB(Register.A0, Register.A0, Register.T0, "Sub operation");
                else if(operator.equals("*"))
                    backend.emitMUL(Register.A0, Register.A0, Register.T0, "Mul operation");
                else if(operator.equals("//"))
                {
                    Label label = generateLocalLabel();
                    backend.emitBNEZ(Register.T0, label, "Check for Divide-by-zero");
                    backend.emitJAL(errorDiv, "Divide-by-zero error");
                    backend.emitLocalLabel(label, "Divide since divisor not zero");
                    backend.emitDIV(Register.A0, Register.A0, Register.T0, "Divide operation");
                }
                else if(operator.equals("%"))
                {
                    Label label = generateLocalLabel();
                    backend.emitBNEZ(Register.T0, label, "Check for Divide-by-zero");
                    backend.emitJAL(errorDiv, "Divide-by-zero error");
                    backend.emitLocalLabel(label, "Divide since divisor not zero");
                    backend.emitREM(Register.A0, Register.A0, Register.T0, "Modulus operation");
                }
                else
                {   // Comparison operators
                    //elseBlock = generateLocalLabel();
                    String comment="Operator: "+operator;
                    if(operator.equals("=="))
                    {   
                        backend.emitXOR(Register.A0, Register.A0, Register.T0, "Check for equality");
                        backend.emitSEQZ(Register.A0, Register.A0, "Result is True if XOR equals 0");
                    }
                    else if(operator.equals("!="))
                    {
                        backend.emitXOR(Register.A0, Register.A0, Register.T0, "Check for inequality");
                        backend.emitSNEZ(Register.A0, Register.A0, "Result is True if XOR does not equal 0");
                    }   
                    else if(operator.equals("<"))
                        backend.emitSLT(Register.A0,Register.A0, Register.T0, comment);
                    else if(operator.equals(">"))
                        backend.emitSLT(Register.A0, Register.T0, Register.A0, comment);
                    else if(operator.equals("<="))
                    {
                        backend.emitADDI(Register.T0,Register.T0, 1, "Increment by 1");
                        backend.emitSLT(Register.A0, Register.A0, Register.T0, comment);
                    }
                    else if(operator.equals(">="))
                    {
                        backend.emitADDI(Register.A0,Register.A0, 1, "Increment by 1");
                        backend.emitSLT(Register.A0, Register.A0, Register.T0, comment);
                    }   
                    else if(operator.equals("is"))
                    {
                        backend.emitXOR(Register.A0,  Register.A0, Register.T0, "is operation");
                        backend.emitSEQZ(Register.A0,  Register.A0, "Result is True if XOR equals 0");
                    }
                    else
                    {}
                }
            }
            else if(node.left.getInferredType().equals(Type.STR_TYPE) && node.right.getInferredType().equals(Type.STR_TYPE))
            {

            }
            return null;
        }
        
        @Override
        public Register analyze(UnaryExpr node) 
        {
            if(node.operator.equals("-") && node.getInferredType().equals(Type.INT_TYPE))
            {
                node.operand.dispatch(this);    
                backend.emitLI(Register.T0, -1, "Set value of Register T0 to -1");
                backend.emitMUL(Register.A0, Register.A0, Register.T0, "Multiply by -1");
            }
            else if(node.operator.equals("not") && node.getInferredType().equals(Type.BOOL_TYPE))
            {    
                node.operand.dispatch(this);
                backend.emitSEQZ(Register.T0, Register.A0, "Not operation on Register A0");
            }
            else
                return null;
            return Register.A0;
        }
        
        @Override
        public Register analyze(Identifier node) 
        {
        		backend.emitLW(T6, FP, 0, "Inside Identifier: ");
            if (sym.getParent() == null) 
            {
                GlobalVarInfo gvi=(GlobalVarInfo) sym.get(node.name);
                backend.emitLW(Register.A0, gvi.getLabel(), "Load global: "+gvi.getVarName());
            }
            else
            {
// FIXME: This breaks, so had to comment it
//            	  VarInfo svi = (VarInfo) sym.get(node.name);
//                int loc = offsetMap.get(svi);
//                backend.emitLW(Register.A0, Register.FP, -loc*4, "Load local variable: "+svi.getVarName());
            }
            return null;
        }
        
        @Override
        public Register analyze(VarDef node) 
        {
            StackVarInfo svi = (StackVarInfo) sym.get(node.var.identifier.name);
            node.value.dispatch(this);
            backend.emitSW(Register.A0, Register.FP, -sp_off*wordSize, "Store variable "+node.var.identifier.name+" value in Stack");
            offsetMap.put(svi, sp_off);
            sp_off++;
            return null;
        }

        @Override
        public Register analyze(WhileStmt node) 
        {
            Label startLoop = generateLocalLabel(); 
            backend.emitLocalLabel(startLoop, "Beginning of while loop");
            node.condition.dispatch(this);
            Label endLoop = elseBlock;
            for(Stmt stmt:node.body)
                stmt.dispatch(this);
            backend.emitJ(startLoop, "Jump to beginning of loop");
            backend.emitLocalLabel(endLoop, "End of while loop");

            return null;
        }
        
        @Override
        public Register analyze(ListExpr node) {
            int l = node.elements.size();
            int i = l;
            for(Expr exp:node.elements)
            {
                Register r = exp.dispatch(this);
                backend.emitSW(r,Register.FP,-sp_off*wordSize,"Push argument "+i+" from last.");
                sp_off++;
                i--;
            }
            backend.emitLI(Register.A0, l, "Pass list length");
            backend.emitSW(Register.A0, Register.FP, -sp_off*wordSize, "Push argument "+i+" from last.");
            sp_off++;
            backend.emitADDI(Register.SP, Register.SP, -sp_off*wordSize, "Set SP to last argument.");
            //TODO: Store reference to variable
            return Register.A0;
        }



        @Override
        public Register analyze(IntegerLiteral node) {
        		backend.emitLW(T6, FP, 0, "Inside IntegerLiteral: " + node.value);
        		backend.emitLI(A0, node.value, "Load integer literal " + node.value);
        		backend.emitJAL(makeIntLabel, "Box integer");
//    			System.out.println("+++ Inside IntegerLiteral");
//    			backend.emitLA(A0, new Label("$int$prototype"), "Load prototype");
//    			backend.emitJAL(new Label("ra, alloc"), "");
//    			backend.emitLI(T0, node.value, "Load integer " + node.value);
//    			backend.emitSW(T0, A0, "@.__int__", null);
        		
            return A0;
        }

        
         @Override
         public Register analyze(ForStmt node) {
            //System.out.println(node);
            /*
            node.
            Label startLoop = generateLocalLabel(); 
            backend.emitLocalLabel(startLoop, "Beginning of while loop");
            node.condition.dispatch(this);
            Label endLoop = elseBlock;
            for(Stmt stmt:node.body)
                stmt.dispatch(this);
            backend.emitJ(startLoop, "Jump to beginning of loop");
            backend.emitLocalLabel(endLoop, "End of while loop");*/
            return null;
            
         }
         
         @Override
         public Register analyze(IndexExpr node) 
         {
            //System.out.println(node);
            return defaultAction(node);
         }
  
        public Register analyze(MemberExpr node) 
        {
            ClassInfo objectClass = (ClassInfo) globalSymbols.get(node.object.getInferredType().className());
            Label label = generateLocalLabel();
            Register obj = node.object.dispatch(this);

            backend.emitBNEZ(obj, label, "Ensure not None");
            backend.emitJ(errorNone, "Go to error handler");
            backend.emitLocalLabel(label, "Not None");
            backend.emitLW(A0, obj, getAttrOffset(objectClass, node.member.name),
                 String.format("Get attribute: %s.%s",  objectClass.getClassName(), node.member.name));
            return A0;
        }
    }

    /**
     * Emits custom code in the CODE segment.
     *
     * <p>This method is called after emitting the top level and the function bodies for each
     * function.
     *
     * <p>You can use this method to emit anything you want outside of the top level or functions,
     * e.g. custom routines that you may want to call from within your code to do common tasks. This
     * is not strictly needed. You might not modify this at all and still complete the assignment.
     *
     * <p>To start you off, here is an implementation of three routines that will be commonly needed
     * from within the code you will generate for statements.
     *
     * <p>The routines are error handlers for operations on None, index out of bounds, and division
     * by zero. They never return to their caller. Just jump to one of these routines to throw an
     * error and exit the program. For example, to throw an OOB error: backend.emitJ(errorOob, "Go
     * to out-of-bounds error and abort");
     */
    protected void emitCustomCode() {
        emitErrorFunc(errorNone, "Operation on None");
        emitErrorFunc(errorDiv, "Division by zero");
        emitErrorFunc(errorOob, "Index out of bounds");
        emitErrorFunc(errorNI, "Not Implemented");
    }

    /** Emit an error routine labeled ERRLABEL that aborts with message MSG. */
    private void emitErrorFunc(Label errLabel, String msg) {
        backend.emitGlobalLabel(errLabel);
        backend.emitLI(A0, ERROR_NONE, "Exit code for: " + msg);
        backend.emitLA(A1, constants.getStrConstant(msg), "Load error message as str");
        backend.emitADDI(
                A1, A1, getAttrOffset(strClass, "__str__"), "Load address of attribute __str__");
        backend.emitJ(abortLabel, "Abort");
    }
}
